/*
 * Structures built up by logic syntax analysis
 */

#include "local.h"
#include "logic.h"

/* object/identifier types; also operators */
#define RESERVED	0
#define SIGNAL		1
#define GROUP		2
#define FUNCTION	3
#define STATE		4
#define BIT		5
#define PARAM		6
#define GRP_ELT		7
#define PIN		8
#define ARRAY		9

/* attribute bits */
#define HIGH_TRUE	1
#define LOW_TRUE	2
#define INPUT_BIT	4
#define OUTPUT_BIT	8
#define TRISTATE	0x10
#define CLOCKED		0x20
#define LATCHED		0x40
#define NO_HAZARD	0x80
#define CRIT_PATH	0x100
#define INTERNAL_SIG	0x200
#define ASYNC_RESET	0x400
#define INPUT_CLOCK	0x800
#define INPUT_LATCH	0x1000

/* the time has come, the Walrus said, to speak of many things */
union object {
	struct signal	*o_sig;		/* if a signal */
	struct expr	*o_grp;		/* if a group name */
	struct function	*o_func;	/* if a function */
	struct state	*o_state;	/* if a state */
	struct expr	*o_ex;		/* if an expression */
	struct param	*o_prm;		/* if a function parameter */
};

/* an identifier */
struct ident {
	char		*i_name;	/* identifier name */
	short		i_type;		/* and type */
	short		i_level;	/* declaration level */
	union object	i_obj;		/* object identified */
	struct ident	*i_next;	/* link in hash table */
};

/* attributes that an input or output signal can have */
struct attr {
	int		a_bits;
	struct expr	*a_clock;
	struct expr	*a_enable;
	struct expr	*a_reset;
	struct expr	*a_inclock;
	struct expr	*a_hazard;
};

/* information about selecting some of a group of objects */
struct index {
	int		i_nel;		/* # elements */
	int		i_minix;	/* lowest index */
	int		i_maxix;	/* highest index */
	int		i_dir;		/* -1 for 3:0, 1 for 0:3 */
};

/* an input or output signal */
struct signal {
	struct attr	s_attr;		/* signal attributes */
	struct index	s_ix;		/* indexing info */
	struct expr	**s_def;	/* ptr to array of defining exprs */
	struct ident	*s_ident;	/* so we know what it's called */
	int		s_num;		/* signal serial number */
};

/* function definition */
struct function {
	int		f_npar;		/* number of parameter bits */
	struct expr	*f_def;		/* defining expression */
};

struct param {
	int		p_num;		/* parameter number */
	struct index	p_ix;		/* indexing info */
};

/* node in expression tree */
struct expr {
	int		e_op;		/* operator */
	union object	e_lu;		/* left expr, leaf, 1st in group */
	union {
		int		er_bval;	/* bit value for BIT */
		int		er_index;	/* index for SIGNAL */
		struct expr	*er_right;	/* right sub-expression */
						/* or last in group */
		struct array	*er_array;	/* logic array */
	} e_ru;
	int		e_nel;		/* # elements in group */
};

#define e_left		e_lu.o_ex
#define e_signal	e_lu.o_sig
#define e_param		e_lu.o_prm
#define e_first		e_lu.o_ex
#define e_state		e_lu.o_state
#define e_right		e_ru.er_right
#define e_last		e_ru.er_right
#define e_index		e_ru.er_index
#define e_bval		e_ru.er_bval
#define e_pin		e_ru.er_index
#define e_array		e_ru.er_array

/* A state of the output variables of a state machine */
struct state {
	struct expr	*s_vars;	/* output vars of this state machine */
	struct expr	*s_value;	/* bit group of output values */
	struct expr	*s_trans;	/* transitions out of this state */
	struct state	*s_next;	/* next state in this machine */
	struct ident	*s_id;		/* so we know its name */
};

struct attr *add_attrs(), *new_attr();
struct ident *new_signal();
struct expr *expr(), *signal_expr(), *bit_expr(), *tt_expr(), *sp_expr();
struct expr *use_func(), *copy_func(), *expand_group(), *empty_group();
struct expr *make_group(), *group_elt(), *out_defs(), *hexnum_expr();
struct expr *pin_expr(), *state_code(), *bits_group(), *ident_expr();
struct expr **def_loc(), *trans_func(), *state_func(), *goto_state();
struct expr *expr_select(), *make_elt();
struct state *find_state();
char *malloc(), *strsave(), *alloc();
long strtol();
int st_outpos();

#define new(x)	((x *)alloc(sizeof(x)))

extern int dec_level;		/* current level of symbol declarations */
extern int num_params;		/* # params so far in function defn */

extern struct expr *inputs;	/* group of inputs seen so far */
extern struct expr *outputs;	/* group of outputs seen so far */
extern int num_signals;		/* total # signals seen so far */

extern int n_errs;		/* # errors encountered */

extern int yylineno;

